// This is a MPI+OpenMP+CUDA program that does the following:
//
// 1. On each host, fill the indices array with the index of each rectangle
// 2. Copy the indices array from each host to its device
// 3. Use the device to run a kernel that:
//    a. Uses the indices array to determine the left x-coordinate of
//       each rectangle (float const x = DeviceIndices[threadId] * width)
//    b. Uses the left x-coordinate to determine the height of the rectangle
//       (float const height = sqrtf(1.0 - x * x)
//    b. Uses the width and height to determine the area
//       (DeviceAreas[threadId] = width * height)
// 4. Copy the areas array from the device to the host
// 5. On the host, sum the areas into a single value
// 6. One process reduces all values and multiplies it by 4 to calculate an
//    approximation of pi
// 6. That process prints the resulting approximation of pi
//
// Author: Aaron Weeden, Shodor, 2016

// Import library so we can call use DBL_EPSILON
#include <float.h>

// Import library so we can call printf()
#include <stdio.h>

// Import library so we can call exit(), malloc(), free(), etc.
#include <stdlib.h>

// Import library with model parameters
#include "params.h"

// Define the number of CUDA threads in each CUDA warp (group of threads that
// execute instructions in lock-step)
#define THREADS_PER_WARP 32

// Define the maximum number of CUDA warps in each CUDA block
#define MAX_WARPS_PER_BLOCK 16

// Define the number of CUDA threads in each CUDA block
#define THREADS_PER_BLOCK ((THREADS_PER_WARP) * (MAX_WARPS_PER_BLOCK))

// Define the number of CUDA blocks in each CUDA grid
#define BLOCKS_PER_GRID 1

// Declare variables for the host and device arrays
int   * HostIndices;
float * HostAreas;
float   Width;
int   * DeviceIndices;
float * DeviceAreas;

// Declare functions that will be defined later
void TryMalloc(void * const err);
void TryCuda(cudaError_t const err);
__global__ void CalcAreas(int const * const indices, float * const areas,
                          float const width, int const count);

// Start the program
int main(int argc, char ** argv)
{
  // Define a variable for the width of a rectangle
  Width = (float)1.0 / RECT_COUNT;

  // Allocate memory for the host arrays
  TryMalloc(HostIndices = (int*)  malloc(INT_BYTE_COUNT));
  TryMalloc(HostAreas   = (float*)malloc(FLOAT_BYTE_COUNT));

  // Allocate memory for the device arrays
  TryCuda(cudaMalloc((void**)&DeviceIndices, INT_BYTE_COUNT));
  TryCuda(cudaMalloc((void**)&DeviceAreas,   FLOAT_BYTE_COUNT));

  // On the host, fill the indices array with the index of each rectangle
  for (int i = 0; i < RECT_COUNT; i++)
  {
    HostIndices[i] = i;
  }

  // Copy the indices array from the host to the device
  TryCuda(cudaMemcpy(DeviceIndices, HostIndices, INT_BYTE_COUNT,
                     cudaMemcpyHostToDevice));

  // Use the device to run a kernel
  CalcAreas<<<BLOCKS_PER_GRID, THREADS_PER_BLOCK>>>(DeviceIndices,
                                                    DeviceAreas,
                                                    Width, RECT_COUNT);

  // Copy the areas array from the device to the host
  TryCuda(cudaMemcpy(HostAreas, DeviceAreas, FLOAT_BYTE_COUNT,
                     cudaMemcpyDeviceToHost));

  // On the host, sum the areas into a single value
  float sum = 0.0;
  for (int i = 0; i < RECT_COUNT; i++)
  {
    sum += HostAreas[i];
  }

  // One process reduces all values and multiplies it by 4 to calculate an
  // approximation of pi
  float pi = 0.0;
  pi = sum;
  pi *= 4.0;

  // That process prints the resulting approximation of pi
  printf("Result:\tpi = %f\n", pi);

  // De-allocate memory for the device arrays
  TryCuda(cudaFree(DeviceAreas));
  TryCuda(cudaFree(DeviceIndices));

  // De-allocate memory for the host arrays
  free(HostAreas);
  free(HostIndices);

  return 0;
}

// Define a function to check whether a malloc() call was successful
void TryMalloc(void * const err)
{
  if (err == NULL)
  {
    fprintf(stderr, "malloc error\n");
    exit(EXIT_FAILURE);
  }
}

// Define a function to check whether a CUDA call was successful
void TryCuda(cudaError_t const err)
{
  if (err != cudaSuccess)
  {
    fprintf(stderr, "CUDA Error: %s\n", cudaGetErrorString(err));
    exit(EXIT_FAILURE);
  }
}

// Define a function which will be executed on a CUDA device
__global__ void CalcAreas(int const * const DeviceIndices,
                          float * const DeviceAreas, float const width,
                          int const count)
{
  // Calculate the unique ID for the current CUDA thread
  int const threadId = blockIdx.x * blockDim.x + threadIdx.x;

  // All threads whose thread ID is >= count will NOT do the following, thus
  // avoiding writing into un-allocated space.
  if (threadId < count)
  {
    float const x = DeviceIndices[threadId] * width;
    float const heightSq = 1.0 - x * x;
    float const height = (heightSq < DBL_EPSILON) ? 0.0 : sqrtf(heightSq);
    DeviceAreas[threadId] = width * height;
  }
}

